# coding: utf-8
"""
Base module for satellite_populate
reads the YAML definition and perform all the rendering and basic actions.
"""
import logging
from collections import Sequence

import fauxfactory
import import_string
from jinja2 import Template
from nailgun import entities
from nailgun.entity_mixins import EntitySearchMixin, EntityReadMixin
# from robottelo.config import settings
from satellite_populate import assertion_operators
from satellite_populate.constants import (
    DEFAULT_CONFIG,
    REQUIRED_MODULES,
    RAW_SEARCH_RULES
)
from satellite_populate.utils import (
    set_logger,
    SmartDict,
    import_from_string,
    format_result,
    remove_keys
)
from six import string_types

logger = logging.getLogger(__name__)


class BasePopulator(object):
    """Base class for API and CLI populators"""

    def __init__(self, data, verbose=None, mode='populate', config=None):
        """Reads YAML and initialize populator"""
        # if not settings.configured:
        #     settings.configure()

        self.data = data
        self.custom_config = config or {}
        self._config = None

        self.mode = mode
        self.vars = SmartDict(data.get('vars', {}))

        self.actions = data.get('actions')
        if not self.actions:
            raise ValueError("Your YAML misses actions: directive")

        self.rendered_actions = []

        self.logger = logger
        set_logger(verbose or 0)

        self.registry = SmartDict()
        self.validation_errors = []
        self.assertion_errors = []
        self.created = []
        self.found = []

        self.context = {'registry': self.registry}
        self.add_modules_to_context()
        self.context.update(self.vars)

        self.load_raw_search_rules()

    # main

    @property
    def config(self):
        """Return config dynamically because it can be overwritten by
        user in datafile or by custom populator"""
        if not self._config:
            self._config = DEFAULT_CONFIG.copy()
            self._config.update(self.custom_config)
            self._config.update(self.data.get('config', {}))
        return self._config

    @property
    def crud_actions(self):
        """Return a list of crud_actions, actions that gets `data`
        and perform nailgun crud operations so custom populators can
        overwrite this list to add new crud actions."""
        return ['create', 'update', 'delete']

    @property
    def raw_search_rules(self):
        """Subclasses of custom populators can extend this rules"""
        return {}

    def execute(self, mode=None):
        """Iterates the entities property described in YAML file
        and parses its values, variables and substitutions
        depending on `mode` execute `populate` or `validate`
        """
        mode = mode or self.mode
        for action_data in self.actions:
            action = action_data.get('action', 'create')

            if action_data.get('when'):
                if not eval(action_data.get('when'), None, self.context):
                    continue

            log_message = action_data['log_message'] = Template(
                action_data.get(
                    'log', action_data.get('register', 'executing...')
                )
            ).render(**self.context)

            getattr(self.logger, action_data.get('level', 'info').lower())(
                '%s: %s',
                action.upper() if not (
                    mode == 'validate' and action == 'create'
                ) else "VALIDATE",
                log_message
            )

            actions_list = self.render(action_data, action)
            for rendered_action_data in actions_list:

                if action not in self.crud_actions:
                    # find the method named as action_name
                    action_name = "action_{0}".format(action.lower())
                    getattr(self, action_name)(
                        rendered_action_data,
                        action_data
                    )

                    self.context.update(self.registry)

                    # check if it is better to do that in method
                    self.add_rendered_action(action_data, rendered_action_data)

                    # execute above method and continue to next item
                    continue

                # Executed only for crud_actions
                model_name = action_data['model'].lower()
                method = getattr(
                    self, '{0}_{1}'.format(mode, model_name), None
                )
                search = self.build_search(
                    rendered_action_data, action_data
                )

                entity_data = remove_keys(
                    rendered_action_data, 'loop_index'
                )

                if method:
                    method(entity_data, action_data, search, action)
                else:
                    # execute self.populate or self.validate
                    getattr(self, mode)(
                        entity_data, action_data, search, action
                    )

                # ensure context is updated with latest created entities
                self.context.update(self.registry)

                self.add_rendered_action(
                    action_data, rendered_action_data
                )

    # renders

    def render(self, action_data, action):
        """Takes an entity description and strips 'data' out to
        perform single rendering and also handle repetitions defined
        in `with_items`
        """

        # TODO: validate required keys marshmallow? voluptuous?
        # if action not in ('delete', 'assertion') and 'data' not
        # in action_data:
        #     raise ValueError('entity misses `data` key')

        items = action_data.get('with_items')
        context = self.context
        entities = []

        if items and isinstance(items, list):
            items = [Template(item).render(**context) for item in items]
        elif items and isinstance(items, string_types):
            items = eval(items, None, context)
            if not isinstance(items, Sequence):
                raise AttributeError("with_items must be sequence type")
        else:
            # as there is no with_items, a single item list will
            # ensure the addition of a single one.
            items = [None, ]

        for loop_index, item in enumerate(items):

            forced_index = action_data.get('loop_index')
            if forced_index is not None and forced_index != loop_index:
                continue

            new_context = {}
            new_context.update(context)
            new_context['item'] = item
            new_context['loop_index'] = loop_index

            # data can be dict on CRUD or list on SPECIAL_ACTIONS
            data = action_data.get('data')
            if isinstance(data, dict):
                rendered_action_data = data.copy()
            else:
                rendered_action_data = {'_values': data}

            rendered_action_data['loop_index'] = loop_index

            self.render_action_data(
                rendered_action_data,
                new_context
            )

            entities.append(rendered_action_data)

        return entities

    def render_action_data(self, data, context):
        """Gets a single action_data and perform inplace template
        rendering or reference evaluation depending on directive being used.
        """
        for k, v in data.items():
            if isinstance(v, dict):
                if 'from_registry' in v:
                    try:
                        if isinstance(v['from_registry'], dict):
                            result = eval(
                                v['from_registry']['name'], None, context
                            )
                            self.resolve_result(
                                data, 'from_registry', k, v, result
                            )
                        else:
                            data[k] = eval(v['from_registry'], None, context)

                    except NameError as e:
                        logger.error(str(e))
                        raise NameError(
                            "{0}: Please check if the reference "
                            "was added to the registry".format(str(e)))
                elif 'from_object' in v:
                    try:
                        if isinstance(v['from_object'], dict):
                            result = import_from_string(
                                v['from_object']['name'])
                            self.resolve_result(
                                data, 'from_object', k, v, result
                            )
                        else:
                            data[k] = import_from_string(v['from_object'])
                    except ImportError as e:
                        logger.error(str(e))
                        raise
                elif 'from_search' in v:
                    try:
                        result = self.from_search(
                            v['from_search'], context
                        )
                        self.resolve_result(
                            data, 'from_search', k, v, result
                        )

                    except Exception as e:
                        logger.error(str(e))
                        raise
                elif 'from_read' in v:
                    try:
                        result = self.from_read(v['from_read'], context)
                        self.resolve_result(
                            data, 'from_read', k, v, result
                        )
                    except Exception as e:
                        logger.error(str(e))
                        raise
                elif 'from_factory' in v:
                    try:
                        data[k] = self.from_factory(v['from_factory'], context)
                    except Exception as e:
                        logger.error(str(e))
                        raise
                else:
                    self.render_action_data(v, context)
            elif isinstance(v, string_types):
                if '{{' in v and '}}' in v:
                    data[k] = Template(v).render(**context)

    # from_ directives

    def from_search(self, action_data, context):
        """Gets fields and perform a search to return Entity object
        used when 'from_search' directive is used in YAML file
        """
        model_name = action_data['model']
        unique = action_data.get('unique', True)
        get_all = action_data.get('all', False)
        index = action_data.get('index', None)
        model = getattr(entities, model_name)

        if 'data' in action_data:
            data = action_data['data'].copy()
            self.render_action_data(data, context)
            search = self.build_search(data, action_data, context)
            search_result = self.get_search_result(
                model, search, unique=unique
            )
        else:
            # empty search returns all when used with filters={}
            options = self.build_search_options({}, action_data)
            self.logger.info(
                "search: getting entities in %s with options %s",
                model_name,
                options
            )
            search_result = model().search(**options)

        silent_errors = action_data.get('silent_errors', False)

        if not search_result:
            self.logger.error("search: returned no objects %s", action_data)
            if silent_errors:
                return None
            raise RuntimeError("Search returned no objects")

        if unique or get_all:
            self.add_to_registry(
                action_data, search_result, append=False)
            return search_result

        self.add_to_registry(
            action_data, search_result[index or 0], append=False)

        return search_result[index or 0]

    def from_read(self, action_data, context):
        """Gets fields and perform a read to return Entity object
        used when 'from_read' directive is used in YAML file
        """

        if 'id' not in action_data['data']:
            raise RuntimeError("read operations demands an id")

        model_name = action_data['model']
        model = getattr(entities, model_name)
        if not issubclass(model, EntityReadMixin):
            raise TypeError("{0} not readable".format(model))

        rendered_action_data = action_data['data'].copy()
        self.render_action_data(rendered_action_data, context)

        entity = model(**rendered_action_data).read()

        self.add_to_registry(action_data, entity, append=False)
        return entity

    def from_factory(self, action_data, context):
        """Generates random content using fauxfactory"""

        method_name = lambda name: "gen_{0}".format(name)  # noqa
        if isinstance(action_data, string_types):
            return getattr(fauxfactory, method_name(action_data))()
        elif isinstance(action_data, dict):
            # only one key is allowed
            if len(action_data.keys()) > 1:
                raise TypeError("from_factory allows only one method name")
            key = action_data.keys()[0]
            value = action_data.values()[0]
            method = getattr(fauxfactory, method_name(key))

            if isinstance(value, dict):
                result = method(**value)
            elif isinstance(value, list):
                result = method(*value)
            else:
                result = method(value)
        else:
            raise TypeError("from_factory receives dict or string")

        self.add_to_registry(action_data, result, append=False)

        return result

    def resolve_result(self, data, from_where, k, v, result):
        """Used in `from_search` and `from_object` to get specific
         attribute from object e.g: name. Or to invoke a method when
         attr is a dictionary of parameters.
        """
        for key_name, value in v[from_where].items():
            # attr. args, key, index
            if key_name == 'attr':
                if isinstance(value, string_types):
                    result = getattr(result, value)
                elif isinstance(value, dict):
                    if len(value.keys()) == 1 and isinstance(
                            value.values()[0], dict):
                        # call named function
                        result = getattr(
                            result, value.keys()[0]
                        )(**value.values()[0])
                    else:
                        # call directly
                        result = result(**value)
                else:
                    raise RuntimeError('attr must be string or dict')
            if key_name == 'args':
                result = result(*value)
            if key_name in ('key', 'index'):
                result = result[value]

        data[k] = result

    # search

    def build_search(self, rendered_action_data, action_data, context=None):
        """Build search data and returns a dict containing elements

        - data
          Dictionary of parsed rendered_action_data to be used to instantiate an object
          to searched without raw_query.

        - options
          if `search_options` are specified it is passed to
          `.search(**options)`

        - searchable
          Returns boolean True if model inherits from EntitySearchMixin, else
          alternative search must be implemented.

        if `search_query` is available in action_data it will be used instead
        of rendered_action_data.
        """

        # if with_items, get current loop_index reference or 0
        loop_index = rendered_action_data.get('loop_index', 0)

        if 'search_query' not in action_data:
            data = rendered_action_data
        else:
            search_data = action_data['search_query']
            if isinstance(search_data, dict):
                items = action_data.get('with_items')

                if items and isinstance(items, list):
                    items = [
                        Template(item).render(**self.context)
                        for item in items
                    ]
                elif items and isinstance(items, string_types):
                    items = eval(items, None, self.context)
                    if not isinstance(items, Sequence):
                        raise AttributeError(
                            "with_items must be sequence type")
                else:
                    # as there is no with_items, a single item list will
                    # ensure the addition of a single one.
                    items = [None, ]

                new_context = {}
                new_context.update(self.context)
                new_context.update(context or {})
                new_context['item'] = items[loop_index]
                new_context['loop_index'] = loop_index
                data = search_data.copy()
                self.render_action_data(data, new_context)

            elif isinstance(search_data, Sequence):
                data = {
                    key: rendered_action_data[key]
                    for key in search_data
                }
            else:
                raise ValueError("search_query bad formatted")

        model_name = action_data['model']
        model = getattr(entities, model_name)

        data = remove_keys(data, 'loop_index')

        options = self.build_search_options(data, action_data)

        # force_raw = action_data.get(
        #     'force_raw', False
        # ) or action_data['model'].lower() in self.force_raw_search

        search = {
            'data': data,
            'options': options,
            'searchable': issubclass(model, EntitySearchMixin)
        }

        return search

    def build_raw_query(self, data, action_data):
        """Builds nailgun raw_query for search"""
        search_data = data.copy()

        rules = self.search_rules.get(action_data['model'].lower())
        if rules:
            for field_name, rule in rules.items():
                if field_name not in search_data:
                    continue

                if rule.get('rename'):
                    value = search_data[field_name]

                    attr = rule.get('attr')
                    index = rule.get('index')
                    key = rule.get('key')

                    if index is not None:
                        value = value[index]

                    if key is not None:
                        value = value[key]

                    if attr:
                        search_data[rule['rename']] = getattr(value, attr)
                    else:
                        search_data[rule['rename']] = value

                if rule.get('remove') or rule.get('rename'):
                    del search_data[field_name]

        query_items = [
            '{0}="{1}"'.format(k, v) for k, v in search_data.items()
        ]
        raw_query = ",".join(query_items)
        return {'search': raw_query or None}

    def build_search_options(self, data, action_data):
        """Builds nailgun options for search
        raw_query:
        Some API endpoints demands a raw_query, so build it as in example:
        `{'query': {'search':'name=name,label=label,id=28'}}`

        force_raw:
        Returns a boolean if action_data.force_raw is explicitly specified

        """
        options = action_data.get('search_options', {})

        force_raw = options.pop(
            'force_raw', False
        ) or action_data['model'].lower() in self.force_raw_search

        if force_raw:
            options['query'] = self.build_raw_query(data, action_data)

        per_page = options.pop('per_page', None)
        if per_page:
            if 'query' in options:
                options['query']['per_page'] = per_page
            else:
                options['query'] = {'per_page': per_page}

        return options

    def get_search_result(self, model, search, unique=False,
                          silent_errors=False):
        """Perform a search"""
        if not search['searchable']:
            self.logger.error('search: %s not searchable', model.__name__)
            if silent_errors:
                return
            raise TypeError("{0} not searchable".format(model.__name__))
        result = model(**search['data']).search(**search['options'])

        options_to_log = search['options'] or {
            k: getattr(v, 'id', v) for k, v in search['data'].items()
        }

        if not result:
            self.logger.info(
                "search: %s %s returned empty result",
                model.__name__,
                options_to_log
            )
            return

        if unique:
            if len(result) > 1:
                self.logger.error(
                    "search: %s %s is not unique",
                    model.__name__,
                    options_to_log
                )
                self.logger.debug(result)
                if not silent_errors:
                    raise RuntimeError(
                        "More than 1 item returned "
                        "search is not unique"
                    )

            self.logger.info(
                "search: %s %s found unique item",
                model.__name__,
                options_to_log,
            )
            return result[0]

        self.logger.info(
            "search: %s %s found %s items",
            model.__name__,
            options_to_log,
            len(result)
        )
        return result

    # helpers

    def load_raw_search_rules(self):
        """Reads default search rules then update first with
        custom populator defined rules and then user defined in datafile.
        """
        self.search_rules = RAW_SEARCH_RULES
        self.search_rules.update(self.raw_search_rules)
        self.search_rules.update(self.config.get('raw_search_rules') or {})

        self.force_raw_search = [
            entity for entity, data in self.search_rules.items()
            if data.get('_force_raw') is True
        ]

    def add_modules_to_context(self):
        """Add modules dynamically to render context"""
        modules_to_add = self.config.get('add_to_context', {})
        modules_to_add.update(REQUIRED_MODULES)
        for name, module in modules_to_add.items():
            self.context[name] = import_string(module)

    def add_to_registry(self, action_data, result, append=True):
        """Add objects to the internal registry"""
        if not action_data.get('register', action_data.get('registry')):
            return

        registry_key = Template(
            action_data.get('register', action_data.get('registry'))
        ).render(**self.context)
        existent = self.registry.get(registry_key)
        is_list = isinstance(existent, list)
        with_items = action_data.get('with_items') is not None

        if not append:
            self.registry[registry_key] = result
            self.logger.info(
                "registry: %s registered as %s",
                format_result(result),
                registry_key
            )
        elif existent and is_list:
            existent.append(result)
            self.logger.info(
                "registry: %s appended to %s",
                format_result(result),
                registry_key
            )
        else:
            self.registry[registry_key] = [result] if with_items else result
            self.logger.info(
                "registry: %s registered as %s",
                format_result(result),
                registry_key
            )

            # if action_data.get('with_items'):
            #     if registry_key in self.registry:
            #         self.registry[registry_key].append(result)
            #         self.logger.info(
            #             "registry: %s appended to %s",
            #             format_result(result),
            #             registry_key
            #         )
            #     else:
            #         self.registry[registry_key] = [result]
            #         self.logger.info(
            #             "registry: %s registered as %s",
            #             format_result(result),
            #             registry_key
            #         )
            # else:
            #     self.registry[registry_key] = result
            #     self.logger.info(
            #         "registry: %s registered as %s",
            #         format_result(result),
            #         registry_key
            #     )

    def add_rendered_action(self, action_data, rendered_action_data):
        """Add rendered action to be written in validation file"""
        data = remove_keys(
            action_data, 'loop_index', '_values', 'log_message',
            deep=True
        )

        def persist_factory(a_dict):
            """Generated data should be persisted in validation file"""
            for k, v in a_dict.items():
                if not isinstance(v, dict):
                    continue
                if 'from_factory' in v:
                    a_dict[k] = rendered_action_data[k]
                else:
                    persist_factory(v)

        # if data not in self.rendered_actions:


        persist_factory(data)
        if data.get('with_items'):
            data['loop_index'] = rendered_action_data['loop_index']

        self.rendered_actions.append(data)

    # special methods

    def action_assertion(self, rendered_action_data, action_data):
        """Run assert operations"""
        data = self.render_assertion_data(
            action_data, rendered_action_data)

        operator = action_data.get('operator', 'eq')
        assertion_function = getattr(assertion_operators, operator)
        assertion_result = assertion_function(**data)

        if assertion_result:
            self.logger.info(
                'assertion: %s is %s to %s',
                data['value'], operator, data['other']
            )
        else:
            self.logger.error(
                'assertion: %s is NOT %s to %s',
                data['value'], operator, data['other']
            )
            self.assertion_errors.append({
                'data': data,
                'operator': operator,
                'action_data': action_data
            })
        self.add_to_registry(action_data, assertion_result)

    def action_echo(self, rendered_action_data, action_data):
        """After message is echoed to log, check if needs print"""
        if action_data.get('print'):
            print(action_data.get('log_message'))  # noqa

    def action_unregister(self, rendered_action_data, action_data):
        """Remove data from registry"""
        for value in rendered_action_data['_values']:
            if self.registry.pop(value, None):
                self.logger.info("unregister: %s OK", value)
            else:
                self.logger.error("unregister: %s IS NOT REGISTERED", value)

    def action_register(self, rendered_action_data, action_data):
        """Register arbitrary items to the registry"""
        for key, value in remove_keys(
                rendered_action_data, 'loop_index').items():
            data = remove_keys(action_data.copy(), 'loop_index')
            data['register'] = key
            self.add_to_registry(data, value)

    # special methods helpers

    def render_assertion_data(self, action_data, rendered_action_data):
        """Render items on assertion data"""
        new_context = self.context.copy()
        new_context['loop_index'] = loop_index = rendered_action_data.get(
            'loop_index')

        with_items = action_data.get('with_items')
        if with_items:
            new_context['item'] = with_items[loop_index]

        data = {
            'value' if index == 0 else 'other': value
            for index, value in enumerate(action_data['data'])
        }
        self.render_action_data(data, new_context)
        return data

    # sub class methods

    def populate(self, rendered_action_data, raw_entity, search_query, action):
        """Should be implemented in sub classes"""
        raise NotImplementedError()

    def validate(self, rendered_action_data, raw_entity, search_query, action):
        """Should be implemented in sub classes"""
        raise NotImplementedError()

    def populate_modelname(self, rendered_action_data, action_data,
                           search_query, action):
        """Example on how to implement custom populate methods
        e.g: `def populate_organization`
        This method should take care of all validations and errors.
        """

        result = (
            "Implement your own populate method here,"
            "This method should take care of `search_query`"
            "to check the existence of entity before creation"
            "and should return a valid Nailgun entity object"
            "containing a valid `id`"
        )

        # always add the result to registry to allow references
        self.add_to_registry(action_data, result)

    def validate_modelname(self, rendered_action_data, action_data,
                           search_query, action):
        """Example on how to implement custom validate methods
        e.g:: `def validate_organization`
        This method should take care of all validations and errors.
        """

        result = (
            "Implement your own validate method here,"
            "This method should use `search_query`"
            "to check the existence of entities"
            "and should add valid Nailgun entity object to self.registry"
            "and errors to self.validation_errors"
        )

        # always add the result to registry to allow references
        self.add_to_registry(action_data, result)

        # add errors to validation_errors
        self.validation_errors.append({
            'search_query': search_query,
            'message': 'entity does not validate in the system',
            'rendered_action_data': rendered_action_data,
            'action_data': action_data
        })



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
test_satellite_populate
----------------------------------

Tests for `satellite_populate` module.
"""

import pytest

from contextlib import contextmanager
from click.testing import CliRunner

from satellite_populate import satellite_populate
from satellite_populate import cli


@pytest.fixture
def response():
    """Sample pytest fixture.
    See more at: http://doc.pytest.org/en/latest/fixture.html
    """
    # import requests
    # return requests.get('https://github.com/audreyr/cookiecutter-pypackage')


def test_content(response):
    """Sample pytest test function with the pytest fixture as an argument.
    """
    # from bs4 import BeautifulSoup
    # assert 'GitHub' in BeautifulSoup(response.content).title.string
def test_command_line_interface():
    runner = CliRunner()
    result = runner.invoke(cli.main)
    assert result.exit_code == 0
    assert 'satellite_populate.cli.main' in result.output
    help_result = runner.invoke(cli.main, ['--help'])
    assert help_result.exit_code == 0
    assert '--help  Show this message and exit.' in help_result.output



# coding: utf-8
"""
This module contains commands to interact with data populator
and validator.

Commands included:

populate
--------

A command to populate the system based in an YAML file describing the
entities::

    $ manage data populate /path/to/file.yaml -vv

validate
--------

A command to validate the system based in an YAML file describing the
entities::

    $ manage data validate /path/to/file.yaml -vv

Use :code:`$ manage data --help` and
:code:`$ manage data populate --help` for more info
"""
import sys
import click
from satellite_populate.main import populate as execute_populate
from satellite_populate.main import validate as execute_validate


@click.group()
def main(args=None):
    """Console script for satellite_populate"""
    click.echo("Replace this message by putting your code into "
               "satellite_populate.cli.main")
    click.echo("See click documentation at http://click.pocoo.org/")


@main.command()
@click.argument('datafile', required=True, type=click.Path())
@click.option('-v', '--verbose', count=True)
@click.option('-o', '--output', type=click.Path(), default=None)
def populate(datafile, verbose, output=None):
    """Populate using the data described in `datafile`:\n
    populated the system with needed entities.\n
        example: $ manage data populate test_data.yaml -vv\n

    verbosity:\n
       -v = DEBUG\n
       -vv = INFO\n
       -vvv = WARNING\n
       -vvvv = ERROR\n
       -vvvvv = CRITICAL\n
    """
    result = execute_populate(datafile, verbose=verbose, output=output)
    result.logger.info(
        "{0} entities already existing in the system".format(
            len(result.found)
        )
    )
    result.logger.info(
        "{0} entities were created in the system".format(
            len(result.created)
        )
    )
    if result.validation_errors:
        for error in result.validation_errors:
            result.logger.error(error['message'])
            result.logger.error(error['search'])


@main.command()
@click.argument('datafile', required=True, type=click.Path())
@click.option('-v', '--verbose', count=True)
def validate(datafile, verbose):
    """Validate using the data described in `datafile`:\n
    populated the system with needed entities.\n
        example: $ manage data populate test_data.yaml -vv\n

    verbosity:\n
       -v = DEBUG\n
       -vv = INFO\n
       -vvv = WARNING\n
       -vvvv = ERROR\n
       -vvvvv = CRITICAL\n

    """

    result = execute_validate(datafile, verbose=verbose)
    if result.assertion_errors:
        for error in result.assertion_errors:
            data = error['data']
            result.logger.error(
                'assertion: %s is NOT %s to %s',
                data['value'], error['operator'], data['other']
            )
        sys.exit("System entities did not validated!")

    if result.validation_errors:
        for error in result.validation_errors:
            result.logger.error(error['message'])
            result.logger.error(error['search'])
        sys.exit("System entities did not validated!")
    else:
        result.logger.info(
            "{0} entities validated in the system".format(
                len(result.found)
            )
        )


if __name__ == "__main__":
    main()



"""This package contains tools to populate and validate the system"""
__author__ = """Bruno Rocha"""
__email__ = 'rochacbruno@gmail.com'
__version__ = '0.1.0'

from satellite_populate.decorators import populate_with  # noqa
from satellite_populate.main import load_data  # noqa



# coding: utf-8
import logging
from copy import deepcopy

import import_string
from nailgun.entity_mixins import Entity
from six import string_types
from six.moves import builtins

try:
    # coloredlogs is optional, used only when installed
    import coloredlogs
except ImportError:
    coloredlogs = None

DICT_RESERVED_KEYS = vars(dict).keys()


class SmartDict(dict):
    """
    A Dict which is accessible via attribute dot notation
    """

    def __init__(self, *args, **kwargs):
        """
        :param args: multiple dicts ({}, {}, ..)
        :param kwargs: arbitrary keys='value'

        If ``keyerror=False`` is passed then not found attributes will
        always return None.
        """
        super(SmartDict, self).__init__()
        self['__keyerror'] = kwargs.pop('keyerror', True)
        [self.update(arg) for arg in args if isinstance(arg, dict)]
        self.update(kwargs)

    def __getattr__(self, attr):
        if attr not in DICT_RESERVED_KEYS:
            if self['__keyerror']:
                return self[attr]
            else:
                return self.get(attr)
        return getattr(self, attr)

    def __setattr__(self, key, value):
        if key in DICT_RESERVED_KEYS:
            raise TypeError("You cannot set a reserved name as attribute")
        self.__setitem__(key, value)

    def __copy__(self):
        return self.__class__(self)

    def copy(self):
        return self.__copy__()


def set_logger(verbose):
    """Set logger verbosity used when client is called with -vvvvv"""

    for _logger in logging.Logger.manager.loggerDict.values():  # noqa
        if not isinstance(_logger, logging.Logger):
            continue
        if coloredlogs is not None:
            _logger.propagate = False

            handler = logging.StreamHandler()
            formatter = coloredlogs.ColoredFormatter(
                fmt="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
                datefmt="%Y-%m-%d %H:%M:%S",
            )
            handler.setFormatter(formatter)
            _logger.handlers = [handler]
        _logger.setLevel(verbose * 10)


def import_from_string(import_name, *args, **kwars):
    """Try import string and then try builtins"""
    try:
        return import_string(import_name, *args, **kwars)
    except:
        import_name = "{0}.{1}".format(builtins.__name__, import_name)
        return import_string(import_name, *args, **kwars)


def format_result(result):
    """format result to show in logs"""
    if isinstance(result, string_types):
        return result[0: 30]
    elif isinstance(result, (list, tuple, dict)):
        return str(result)[0: 30]
    elif isinstance(result, Entity):
        return "{0}:{1}".format(result.__class__.__name__, result.id)
    else:
        return result


def remove_keys(data, *args, **kwargs):
    """remove keys from dictionary
    d = {'item': 1, 'other': 2, 'keep': 3}
    remove_keys(d, 'item', 'other')
    d -> {'keep': 3}
    deep = True returns a deep copy of data.
    """
    if kwargs.get('deep') is True:
        data = deepcopy(data)

    return {
        k: v for k, v in data.items() if k not in args
    }



"""To be implemented: a populator using CLI"""

# from satellite_populate.base import BasePopulator
#
#
# class CLIPopulator(BasePopulator):
#     """Populates system using hammer"""
#     # TO BE IMPLEMENTED



# coding: utf8
"""Point of entry for populate and validate used in scripts"""
import json
from collections import OrderedDict

import import_string
import os
import yaml
from satellite_populate.constants import DEFAULT_CONFIG


def setup_yaml():
    """Set YAML to use OrderedDict
    http://stackoverflow.com/a/8661021"""
    represent_dict_order = lambda self, data: self.represent_mapping(  # noqa
        'tag:yaml.org,2002:map', data.items()
    )
    yaml.add_representer(OrderedDict, represent_dict_order)


setup_yaml()


def load_data(datafile):
    """Loads YAML file as a dictionary"""
    if datafile.endswith(('.yml', '.yaml', 'json')):
        with open(datafile) as datafile:
            return yaml.load(datafile)
    return yaml.load(datafile)


def get_populator(data, verbose):
    """Gets an instance of populator dynamically"""
    if not isinstance(data, dict):
        data = load_data(data)

    config = DEFAULT_CONFIG.copy()
    config.update(data.get('config', {}))
    verbose = verbose or config.get('verbose')

    populator_name = config['populator']
    populator_module_name = config['populators'][populator_name]['module']
    populator_class = import_string(populator_module_name)
    populator = populator_class(data=data, verbose=verbose, config=config)
    return populator


def populate(data, **extra_options):
    """Loads and execute populator in populate mode"""
    verbose = extra_options.get('verbose')
    populator = get_populator(data, verbose)
    populator.execute(mode='populate')
    populator.logger.info("Populator finished!")
    if extra_options.get('output'):
        save_rendered_data(populator, extra_options['output'])
    return populator


def validate(data, **extra_options):
    """Loads and execute populator in validate mode"""
    verbose = extra_options.get('verbose')
    populator = get_populator(data, verbose)
    populator.execute(mode='validate')
    populator.logger.info("Validator finished!")
    if extra_options.get('output'):
        save_rendered_data(populator, extra_options['output'])
    return populator


def wrap_context(result):
    """Takes the result of populator and keeps only useful data
    e.g. in decorators context.registered_name, context.config.verbose and
    context.vars.admin_username will all be available.
    """
    context = result.registry
    context.config = result.config
    context.vars = result.vars
    return context


def save_rendered_data(result, filename):
    """Save the result of rendering in a new file to be used for
    validation"""
    file_format = filename.rsplit('.')[-1]
    if file_format not in ('json', 'yml', 'yaml', 'py'):
        raise ValueError("Invalid filename extension")

    data = OrderedDict({
        'config': dict(result.config),
        'vars': dict(result.vars),
        'actions': result.rendered_actions
    })

    if not filename.startswith('validation_'):
        filename = "validation_{0}".format(filename)

    with open(filename, 'w') as output:
        if file_format in ('yml', 'yaml'):
            output.write(yaml.dump(data))
        elif file_format == 'json':
            output.write(json.dump(data))
        elif file_format == 'py':
            output.write(str(data))

    result.logger.debug("Validation data saved in %s", filename)



# coding: utf-8
"""Default base config values"""
from satellite_populate.utils import SmartDict

REQUIRED_MODULES = {
    'fauxfactory': 'fauxfactory',
    'env': 'os.environ'
}

RAW_SEARCH_RULES = {
    # force Organization to always perform raw_search
    'organization': {'_force_raw': True},
    'user': {
        '_force_raw': True,
        'organization': {
            # rename organization Entity to organization_id
            'rename': 'organization_id',
            # and get the attr 'id' from object
            'attr': 'id',
            # using object in index 0 (because it is a list)
            'index': 0,
            # if was dict key_name here
            # 'key': 'name_of_key'
        },
        # remove fields from search
        'password': {'remove': True},
        'default_organization': {'remove': True}
    },
    'repository': {
        'url': {'remove': True},
    }
}

DEFAULT_CONFIG = SmartDict({
    'populator': 'api',
    'populators': {
        'api': {
            'module': 'satellite_populate.api.APIPopulator'
        }
    },
    'verbose': 0
})



# coding: utf-8
"""
Implement basic assertions to be used in assertion action
"""


def eq(value, other):
    """Equal"""
    return value == other


def ne(value, other):
    """Not equal"""
    return value != other


def gt(value, other):
    """Greater than"""
    return value > other


def lt(value, other):
    """Lower than"""
    return value < other


def gte(value, other):
    """Greater than or equal"""
    return value >= other


def lte(value, other):
    """Lower than or equal"""
    return value <= other


def identity(value, other):
    """Identity check using ID"""
    return value is other



"""decorators for populate feature"""
from functools import wraps

from satellite_populate.main import populate, wrap_context


def populate_with(data, context=None, **extra_options):
    """To be used in test cases as a decorator

    Having a data_file like::

        actions:
          - model: Organization
            register: organization_1
            data:
              name: My Org

    Then you can use in decorators::

        @populate_with('file.yaml')
        def test_case_(self):
            'My Org exists in system test anything here'

    And getting the populated entities inside the test_case::

        @populate_with('file.yaml', context=True)
        def test_case_(self, context=None):
            assert context.entities.organization_1.name == 'My Org'

    You can also set a name to the context argument::

        @populate_with('file.yaml', context='my_context')
        def test_case_(self, my_context=None):
            assert my_context.organization_1.name == 'My Org'

    NOTE::

        That is important that ``context`` argument always be declared using
        either a default value ``my_context=None`` or handle in ``**kwargs``
        Otherwise ``py.test`` may try to use this as a fixture placeholder.

    """

    def decorator(func):
        """Wrap test method"""

        @wraps(func)
        def wrapper(*args, **kwargs):
            """decorator wrapper"""

            result = populate(data, **extra_options)
            if context:
                context_name = context if context is not True else 'context'
                kwargs[context_name] = wrap_context(result)

            return func(*args, **kwargs)

        return wrapper

    return decorator



# coding: utf-8
"""Implements API populator using Nailgun"""
from nailgun import entities
from nailgun.entity_mixins import EntitySearchMixin
from requests.exceptions import HTTPError
from satellite_populate.base import BasePopulator


class APIPopulator(BasePopulator):
    """Populates system using API/Nailgun"""

    def add_and_log_error(self, action_data, rendered_action_data,
                          search, e=None):
        """Add to validation errors and outputs error"""
        error_message = "%s: %s %s" % (
            self.mode, str(e) if e else '', action_data)
        if e and hasattr(e, 'response'):
            error_message += str(e.response.content)
        self.logger.error(error_message)
        self.validation_errors.append({
            'search': search,
            'message': error_message,
            'rendered_action_data': rendered_action_data,
            'action_data': action_data
        })

    def populate(self, rendered_action_data, action_data, search, action):
        """Populates the System using Nailgun
        based on value provided in `action` argument gets the
        proper CRUD method to execute dynamically
        """
        model = getattr(entities, action_data['model'])
        silent_errors = action_data.get('silent_errors', False)

        try:
            # self.action_create|action_update|action_delete
            result = getattr(self, "action_{0}".format(action))(
                rendered_action_data, action_data, search, model, silent_errors
            )
        except HTTPError as e:
            self.add_and_log_error(action_data, rendered_action_data,
                                   search, e)
        except Exception as e:
            self.add_and_log_error(action_data, rendered_action_data,
                                   search, e)
            if silent_errors:
                return
            raise
        else:
            self.add_to_registry(action_data, result)

    def action_create(self, rendered_action_data, action_data, search, model,
                      silent_errors):
        """Creates new entity if does not exists or get existing
        entity and return Entity object"""
        result = self.get_search_result(
            model, search, unique=True, silent_errors=silent_errors
        )
        if result:
            self.logger.info(
                "create: Entity already exists: %s %s",
                model.__name__,
                result.id
            )
            self.found.append(result)
        else:
            result = model(**rendered_action_data).create()
            self.logger.info(
                "create: Entity created: %s %s", model.__name__, result.id
            )
            self.created.append(result)
        return result

    def action_update(self, rendered_action_data, action_data, search, model,
                      silent_errors):
        """Updates an existing entity"""

        search_data = action_data.get('search_query')
        entity_id = action_data.get('id')

        if not entity_id and not search_data:
            raise RuntimeError("update: missing id or search_query")

        if not entity_id:
            search_result = self.get_search_result(
                model, search, unique=True, silent_errors=silent_errors
            )
            if not search_result:
                raise RuntimeError("update: Cannot find entity")

            entity_id = search_result.id

        entity = model(id=entity_id, **rendered_action_data)
        entity.update(rendered_action_data.keys())
        self.logger.info("update: %s %s", model, entity_id)
        return entity

    def action_delete(self, rendered_action_data, action_data, search, model,
                      silent_errors):
        """Deletes an existing entity"""

        search_data = action_data.get('search_query')
        entity_id = action_data.get('id')

        if not entity_id and not search_data:
            raise RuntimeError("delete: missing id or search_query")

        if not entity_id:
            search_result = self.get_search_result(
                model, search, unique=True, silent_errors=silent_errors
            )
            if not search_result:
                raise RuntimeError("delete: Cannot find entity")

            entity_id = search_result.id

        # currently only works based on a single id
        # should iterate all results and delete one by one?
        model(id=entity_id).delete()
        self.logger.info("delete: %s %s", model, entity_id)

    def validate(self, rendered_action_data, action_data, search, action):
        """Based on action fields or using action_data['search_query']
        searches the system and validates the existence of all entities
        """
        silent_errors = action_data.get('silent_errors', False)
        if action not in ['create', 'assertion'] or action_data.get(
                'skip_validation'):
            # validate only create and assertion crud actions
            return

        model = getattr(entities, action_data['model'])

        if not issubclass(model, EntitySearchMixin):
            raise TypeError("{0} not searchable".format(model.__name__))

        try:
            # 1) check if entity exists
            result = self.get_search_result(
                model, search, unique=True, silent_errors=silent_errors
            )
        except HTTPError as e:
            self.add_and_log_error(action_data, rendered_action_data,
                                   search, e)
        else:
            if result:
                self.logger.info(
                    "validate: Entity found: %s %s", model.__name__, result.id
                )
                self.found.append(result)
                self.add_to_registry(action_data, result)
            else:
                # should add None to registry, else references fail
                self.logger.info(
                    "validate: result not found for query %s %s",
                    model.__name__,
                    action_data
                )
                self.add_to_registry(action_data, None)
                self.add_and_log_error(
                    action_data, rendered_action_data,
                    search, 'entity does not exist in the system'
                )
